from itertools import product
import numpy as np
import pandas as pd
import pytest
from .._utils import (
    _check_indices,
    format_table,
    make_bold,
    make_mean_sem_table,
    export_longtable,
)

classifiers = {"KNN": "K-NN", "LR": "LogReg", "DT": "DecTree"}
metrics = {"F1": "F-Score", "GM": "G-mean", "OA": "Acc"}
columns = ["Prop", "Bench2", "Bench1"]

rng = np.random.default_rng(42)

index = np.array(
    list(
        product(
            list(classifiers.keys()),
            list(metrics.keys()),
        )
    )
)
rng.shuffle(index)
rng.shuffle(columns)
X = rng.random((9, 3))

table = pd.DataFrame(
    data=X,
    index=pd.MultiIndex.from_arrays(index.T, names=["Clf", "Perf"]),
    columns=columns,
)

indices_list = [
    None,
    list(classifiers.keys()),
    [list(classifiers.keys()), list(metrics.keys())],
    [classifiers, metrics],
    classifiers,
    {"Perf": metrics},
    {"Perf": list(metrics.keys()), "Clf": list(classifiers.keys())},
    {"Perf": metrics, "Clf": classifiers},
]


@pytest.mark.parametrize("indices", indices_list)
def test_check_indices(indices):
    none_indices = {
        "Clf": {"LR": "LR", "KNN": "KNN", "DT": "DT"},
        "Perf": {"F1": "F1", "GM": "GM", "OA": "OA"},
    }
    indices_ = _check_indices(table.index, indices)

    assert type(indices_) == dict

    if indices is None:
        assert indices_ == none_indices

    if indices != {"Perf": metrics}:
        assert "Clf" in indices_.keys()
    else:
        assert "Perf" in indices_.keys()


@pytest.mark.parametrize(
    "columns", [{"Prop": "a", "Bench1": "b"}, ["Prop"], ["Prop", "Bench2", "Bench1"]]
)
def test_format_table(columns):

    indices = {"Perf": metrics}

    tab = format_table(table, indices=indices, columns=columns, drop_missing=False)
    assert len(tab.columns == 3)

    table_no_index = table.reset_index(drop=True)
    indices = list(range(table_no_index.shape[0]))
    np.random.shuffle(indices)

    tab = format_table(
        table_no_index, indices=indices[:-1], columns=columns, drop_missing=True
    )
    assert tab.index.tolist() == indices[:-1]
    assert tab.shape == (len(indices) - 1, len(columns))
    assert tab.index.tolist() == indices[:-1]


def test_make_bold():

    # Default values
    for i in [0, 1]:
        exp_max_indices = np.argmax(table.values, axis=i)
        table_bf = make_bold(table, decimals=5, axis=i)
        max_indices = np.argmax(
            table_bf.applymap(lambda x: x.startswith("\\textbf")).values, axis=i
        )
        assert (max_indices == exp_max_indices).all()

    # Threshold - higher than
    bold_test = [
        (table.values > 0.5, True),
        (table.values < 0.5, False),
    ]
    for exp_is_bold, maximum in bold_test:
        table_bf = make_bold(table, decimals=5, maximum=maximum, threshold=0.5)
        is_bold = table_bf.applymap(lambda x: x.startswith("\\textbf")).values
        assert (is_bold == exp_is_bold).all()


@pytest.mark.parametrize(
    "sem_vals, make_bold", product([None, rng.random((9, 3))], [True, False])
)
def test_make_mean_sem_table(sem_vals, make_bold):

    exp_max_indices = np.argmax(table.values, axis=1)
    mean_sem = make_mean_sem_table(table, sem_vals=sem_vals, make_bold=make_bold)

    if make_bold:
        max_indices = np.argmax(
            mean_sem.applymap(lambda x: x.startswith("\\textbf")).values, axis=1
        )
        assert (max_indices == exp_max_indices).all()

    if sem_vals is not None:
        assert mean_sem.applymap(lambda x: "$\\pm$" in x).values.all()

    if not make_bold and sem_vals is None:
        assert (mean_sem.values == table.round(2).astype(str).values).all()


def test_export_latex_longtable():
    mean_sem = make_mean_sem_table(table, rng.random((9, 3)), make_bold=True)
    longtable = export_longtable(mean_sem, index=True)
    assert longtable.startswith("\\begin{longtable}{ccc}")
    assert longtable.endswith("\\end{longtable}\n")
    assert longtable.count("$\\pm$") == mean_sem.size



"""
Functions for visualization formatting or producing pre-formatted
visualizations.
"""
from distutils.spawn import find_executable
import warnings
import types
import numpy as np


def _optional_import(module: str) -> types.ModuleType:
    """
    Import an optional dependency.

    Parameters
    ----------
    module : str
        The identifier for the backend. Either an entrypoint item registered
        with importlib.metadata, "matplotlib", or a module name.

    Returns
    -------
    types.ModuleType
        The imported backend.
    """
    # This function was adapted from the _load_backend function from the pandas.plotting
    # source code.
    import importlib

    # Attempt an import of an optional dependency here and raise an ImportError if
    # needed.
    try:
        module_ = importlib.import_module(module)
    except ImportError:
        mod = module.split(".")[0]
        raise ImportError(f"{mod} is required to use this functionality.") from None

    return module_


def set_matplotlib_style(font_size=8, **rcparams):
    """
    Load LaTeX-style configurations for Matplotlib Visualizations.
    """
    plt = _optional_import("matplotlib.pyplot")

    # Replicates the rcParams of seaborn's "whitegrid" style and a few extra
    # configurations I like
    plt.style.use("seaborn-v0_8-whitegrid")
    base_style = {
        # "patch.edgecolor": "w",
        # "patch.force_edgecolor": True,
        # "xtick.bottom": False,
        # "ytick.left": False,
        "font.family": "serif",
        # Use 10pt font in plots, to match 10pt font in document
        "axes.labelsize": (10 / 8) * font_size,
        "font.size": (10 / 8) * font_size,
        # Make the legend/label fonts a little smaller
        "legend.fontsize": font_size,
        "xtick.labelsize": font_size,
        "ytick.labelsize": font_size,
        # Subplots size/shape
        "figure.subplot.left": 0.098,
        "figure.subplot.right": 0.938,
        "figure.subplot.bottom": 0.12,
        "figure.subplot.top": 0.944,
        "figure.subplot.wspace": 0.071,
    }
    plt.rcParams.update(base_style)

    if find_executable("latex"):
        tex_fonts = {
            # Use LaTeX to write all text
            "text.usetex": True,
        }
        plt.rcParams.update(tex_fonts)
    else:
        warn_msg = (
            "Could not find a LaTeX installation. ``text.usetex`` will be set to False."
        )
        warnings.warn(warn_msg)

    # Used to pass any additional custom configurations
    plt.rcParams.update(rcparams)


def feature_to_color(col, cmap="RdYlBu_r"):
    """
    Converts a column of values to hex-type colors.

    Parameters
    ----------
    col : {list, array-like} of shape (n_samples,)
        Values to convert to hex-type color code

    cmap : str or `~matplotlib.colors.Colormap`
        The colormap used to map normalized data values to RGBA colors

    Returns
    -------
    colors : array-like of shape (n_samples,)
        Array with hex values as string type.
    """
    colors = _optional_import("matplotlib.colors")
    cm = _optional_import("matplotlib.cm")

    if type(col) == list:
        col = np.array(col)

    norm = colors.Normalize(vmin=col.min(), vmax=col.max(), clip=True)
    mapper = cm.ScalarMappable(norm=norm, cmap=cmap)
    rgba = mapper.to_rgba(col)

    return np.apply_along_axis(colors.rgb2hex, 1, rgba)



import numpy as np
import pandas as pd
from scipy.sparse import issparse
from sklearn.base import clone, TransformerMixin
from sklearn.utils.metaestimators import _BaseComposition
from sklearn.preprocessing._encoders import OneHotEncoder


class PipelineEncoder(TransformerMixin, _BaseComposition):
    """
    Pipeline-compatible wrapper of Scikit-learn's Transformer objects. Used to pass
    encoding of non-metric features and scalers (when there are categorical features)
    within a pipeline.

    The fitted encoder object from Scikit-learn is stored in ``self.encoder_``.

    .. warning::
        In most cases, ``sklearn.compose.ColumnTransformer`` should be used instead of
        ``PipelineEncoder``. This object might be removed in the future.

    Parameters
    ----------
    features : ndarray of shape (n_cat_features,) or (n_features,)
        Specifies which features to transform. Can either be:

        - array of indices specifying the features to transform.
        - mask array of shape (n_features, ) and ``bool`` dtype for which
          ``True`` indicates the features to transform.
        - array of shape (n_transf_features,) and ``str`` dtype with the names of the
          features to transform. In this case, ``X`` must be a dataframe. Raises an
          error otherwise.

    encoder : encoder object, default=None
        Encoder object to be used for transforming the features. If None,
        defaults to sklearn's OneHotEncoder with default parameters, which can be
        modified with keyword arguments.

        .. warning::
            The ``encoder`` object must be compatible with sklearn's API.

    **kwargs : dict
        Parameters to define the ``OneHotEncoder`` when ``encoder`` is None. Otherwise,
        it is ignored.

    Attributes
    ----------
    features_ : array of shape (n_features,)
        Mask array of shape (n_features, ) and ``bool`` dtype for which
        ``True`` indicates the features to transform.
    """

    _estimator_type = "encoder"

    def __init__(self, features=None, encoder=None, **kwargs):
        self.features = features
        self.encoder = encoder
        self._kwargs = kwargs

    def _check_X(self, X):
        """
        Perform custom check_array. Collect information regarding the passed data and
        ensure the data is a numpy array.
        """
        if type(X) == pd.DataFrame:
            self._is_pandas = (
                True if not hasattr(self, "_is_pandas") else self._is_pandas
            )
            self.columns_ = X.columns
            X_ = X.copy().values
        else:
            self._is_pandas = (
                False if not hasattr(self, "_is_pandas") else self._is_pandas
            )
            X_ = X.copy()
        return X_

    def _check_features(self, X):
        """
        Preprocess ``features``. Converts ``features`` to a mask
        array.
        """
        if self.features is None or len(self.features) == 0:
            cat_features = np.zeros(X.shape[-1]).astype(bool)

        elif type(self.features) in [str, bool, int, float]:
            cat_features = np.array([self.features])

        elif hasattr(self.features, "__iter__"):
            if len(set([type(i) for i in self.features])) != 1:
                raise TypeError(
                    "``features`` cannot have more than one type of object."
                )
            cat_features = np.array(self.features)

        else:
            error_msg = (
                "``features`` must be an iterable or one of str,"
                + " bool, int, float or NoneType. Got "
                + f"{type(self.features).__name__} instead."
            )
            raise TypeError(error_msg)

        is_mask = np.array([type(i) == np.bool_ for i in cat_features]).all()
        is_indices = np.array(
            [type(i) in [np.int64, np.float64] for i in cat_features]
        ).all()
        is_col_names = np.array([type(i) == np.str_ for i in cat_features]).all()

        if is_mask:
            return cat_features

        elif is_indices:
            features_ = np.zeros(X.shape[-1])
            features_[cat_features] = 1
            return features_.astype(bool)

        elif is_col_names:
            if not self._is_pandas:
                error_msg = (
                    "If ``features`` contains string values, "
                    + "``X`` must be a pandas dataframe."
                )
                raise TypeError(error_msg)
            in_columns = np.array([col in X.columns for col in cat_features])
            if any(~in_columns):
                not_in_columns = np.array(cat_features)[np.where(~in_columns)[0]]
                raise KeyError(", ".join(not_in_columns))
            elif self._is_pandas and any(X.columns.isin(cat_features)):
                return X.columns.isin(cat_features)

        else:
            raise TypeError(
                "Could not parse which features to transform from "
                + f"``features``. Got {self.features}."
            )

    def fit(self, X, y=None):
        """
        Fit PipelineEncoder to X.

        Parameters
        ----------
        X : array-like of shape (n_samples, n_features)
            The data to determine the categories of each feature.

        y : None
            Ignored. This parameter exists only for compatibility with
            :class:`~sklearn.pipeline.Pipeline`.

        Returns
        -------
        self
            Fitted encoder.
        """

        X_ = self._check_X(X)

        self.features_ = self._check_features(X)

        if not self.features_.any():
            # If there are no features apply no change
            return self

        self.encoder_ = (
            clone(self.encoder)
            if self.encoder is not None
            else OneHotEncoder(**self._kwargs)
        )
        self.encoder_.fit(X_[:, self.features_], y)

        return self

    def transform(self, X):
        """
        Transform X using the ``encoder`` object.

        Parameters
        ----------
        X : array-like of shape (n_samples, n_features_to_encode + n_remaining)
            Data containing the features to encode.

        Returns
        -------
        X_out : {ndarray, sparse matrix} of shape \
                (n_samples, n_encoded_features + n_remaining)
            Transformed input. Regardless of `sparse_output`, a dense matrix will be
            returned.
        """

        X_ = self._check_X(X)

        if not self.features_.any():
            # If there are no features apply no change
            return X_

        if self._is_pandas:
            metric_data = pd.DataFrame(
                X_[:, ~self.features_],
                columns=self.columns_[~self.features_],
            )
            enc_vals = self.encoder_.transform(X_[:, self.features_])
            encoded_data = pd.DataFrame(
                enc_vals if not issparse(enc_vals) else enc_vals.toarray(),
                columns=self.encoder_.get_feature_names_out(
                    self.columns_[self.features_]
                ),
            )
            data = pd.concat([metric_data, encoded_data], axis=1)

        else:
            metric_data = X_[:, ~self.features_]
            enc_vals = self.encoder_.transform(X_[:, self.features_])
            encoded_data = enc_vals if not issparse(enc_vals) else enc_vals.toarray()
            data = np.concatenate([metric_data, encoded_data], axis=1).astype(
                np.float64
            )

        return data



import numpy as np
from sklearn.metrics import make_scorer
from sklearn.metrics._scorer import _PredictScorer, _SCORERS
from imblearn.metrics import geometric_mean_score


class ALScorer(_PredictScorer):
    """
    Make an Active Learning scorer from a AL-specific metric or loss function.

    This factory class wraps scoring functions to be used in
    :class:`~rlearn.model_selection.ModelSearchCV` and
    :class:`~sklearn.model_selection.GridSearchCV`. It takes a score function, such as
    :func:`~research.metrics.area_under_learning_curve` or
    :func:`~research.metrics.data_utilization_rate` and is used to score an AL
    simulation. The signature of the call is `(estimator, X, y)` where `estimator` is the
    model to be evaluated, `X` is the data and `y` is the ground truth labeling (or
    `None` in the case of unsupervised models).

    Parameters
    ----------
    score_func : callable
        Score function (or loss function) with signature
        ``score_func(y, y_pred, **kwargs)``.

    sign : int, default=1
        Use 1 to keep the original variable's scale, use -1 to reverse the scale.

    Returns
    -------
    scorer : callable
        Callable object that returns a scalar score.
    """

    def __init__(self, score_func):
        super().__init__(score_func=score_func, sign=1, kwargs={})

    def _score(self, method_caller, estimator, X, y_true, sample_weight=None):
        """Evaluate predicted target values for X relative to y_true.
        Parameters
        ----------
        method_caller : callable
            Returns predictions given an estimator, method name, and other
            arguments, potentially caching results.
        estimator : object
            Trained estimator to use for scoring. Must have a predict_proba
            method; the output of that is used to compute the score.
        X : {array-like, sparse matrix}
            Test data that will be fed to estimator.predict.
        y_true : array-like
            Gold standard target values for X.
        sample_weight : array-like of shape (n_samples,), default=None
            Sample weights.

        Returns
        -------
        score : float
            Score function applied to prediction of estimator on X.
        """

        metadata = estimator.metadata_

        return self._sign * self._score_func(metadata)


def geometric_mean_score_macro(y_true, y_pred):
    """Geometric mean score with macro average."""
    return geometric_mean_score(y_true, y_pred, average="macro")


def area_under_learning_curve(metadata, *args):
    """Area under the learning curve. Used in Active Learning experiments."""
    iterations = np.sort([i for i in metadata.keys() if type(i) == int])[1:]
    test_scores = [metadata[i]["test_score"] for i in iterations]
    auc = np.sum(test_scores) / len(test_scores)
    return auc


def data_utilization_rate(metadata, threshold=0.8):
    """Data Utilization Rate. Used in Active Learning Experiments."""
    iterations = np.sort([i for i in metadata.keys() if type(i) == int])[1:]
    test_scores = [metadata[i]["test_score"] for i in iterations]
    n_obs = metadata["data"][0].shape[0]
    data_utilization = [
        metadata[i - 1]["labeled_pool"].sum() / n_obs for i in iterations
    ]

    indices = np.where(np.array(test_scores) >= threshold)[0]
    arg = indices[0] if len(indices) != 0 else -1
    dur = data_utilization[arg] if arg != -1 else np.nan
    return dur


_SCORERS["geometric_mean_score_macro"] = make_scorer(geometric_mean_score_macro)
_SCORERS["area_under_learning_curve"] = ALScorer(area_under_learning_curve)
_SCORERS["data_utilization_rate"] = ALScorer(data_utilization_rate)



from os import remove
from pandas import DataFrame
from .._data import load_datasets
from ...datasets import ContinuousCategoricalDatasets


def test_load_datasets():
    DATASET_NAMES = ["echocardiogram", "acute"]

    # load and save datasets
    datasets = ContinuousCategoricalDatasets(names=DATASET_NAMES)
    datasets.download().save(".", "sample_data_base")
    for name, data in datasets.content_:
        data.to_csv(f"{name.lower()}_data_file.csv", index=False)

    # test csv passing suffix
    csv_datasets = load_datasets(".", suffix=".csv")

    # test .db passing prefix
    db_datasets = load_datasets(".", prefix="sample_")

    # load all (no filters) without target
    all_datasets = load_datasets(".", target_exists=False)

    # delete data files
    files = [f"{name}_data_file.csv" for name in DATASET_NAMES] + [
        "sample_data_base.db"
    ]
    for file in files:
        remove(file)

    exp_csv_datasets = [f"{name.upper()} DATA FILE" for name in DATASET_NAMES]
    exp_db_datasets = [name.upper() for name in DATASET_NAMES]
    exp_all_datasets = [*exp_db_datasets, *exp_csv_datasets]

    assert len(csv_datasets) == len(DATASET_NAMES)
    assert len(db_datasets) == len(DATASET_NAMES)
    assert len(all_datasets) == len(DATASET_NAMES) * 2

    assert sorted(list(dict(csv_datasets).keys())) == sorted(exp_csv_datasets)
    assert sorted(list(dict(db_datasets).keys())) == sorted(exp_db_datasets)
    assert sorted(list(dict(all_datasets).keys())) == sorted(exp_all_datasets)

    assert [len(dat[-1]) for dat in csv_datasets] == [2, 2]
    assert [len(dat[-1]) for dat in db_datasets] == [2, 2]
    assert [type(dat[-1]) for dat in all_datasets] == [DataFrame for i in range(4)]



"""
Base class for Active Learning models
"""

# Author: Joao Fonseca <jpfonseca@novaims.unl.pt>
# License: MIT

from abc import ABCMeta, abstractmethod
from typing import Union
import numpy as np
from sklearn.base import ClassifierMixin, BaseEstimator, clone
from sklearn.utils import check_X_y
from sklearn.ensemble import RandomForestClassifier

from ._acquisition_functions import ACQUISITION_FUNCTIONS
from ..metrics import get_scorer


class BaseActiveLearner(BaseEstimator, metaclass=ABCMeta):
    """
    Base class to implement Active Learning models.

    Parameters
    ----------
    classifier : classifier object, default=None
        Classifier or pipeline to be trained in the iterative process. If None, defaults
        to sklearn's RandomForestClassifier with default parameters and uses the
        ``random_state`` passed in the Active Learning model.

    acquisition_func : function or {'entropy', 'breaking_ties',\
        'random'}, default=None
        Method used to quantify the prediction's uncertainty level. All predefined
        functions are set up so that a higher value means higher uncertainty (higher
        likelihood of selection) and vice-versa. The uncertainty estimate is used to
        select the instances to be added to the labeled/training dataset. Acquisition
        functions may be added or changed in the ``UNCERTAINTY_FUNCTIONS`` dictionary.
        If None, defaults to "random".

    n_init : int or float, default=None
        Number of observations to include in the initial training dataset. If
        ``n_init`` < 1, then the corresponding percentage of the original dataset
        will be used as the initial training set. If None, defaults to 2% of the size of
        the original dataset.

    budget : int or float, default=None
        Number of observations to be added to the training dataset at each iteration. If
        ``budget`` < 1, then the corresponding percentage of the original dataset will be
        used as the initial training set. If None, defaults to 2% of the size of the
        original dataset.

    max_iter : int, default=None
        Maximum number of iterations allowed. If None, the experiment will run until 100%
        of the dataset is added to the training set.

    evaluation_metric : string, default='accuracy'
        Metric used to calculate the test scores. See
        ``research.metrics`` for info on available
        performance metrics.

    continue_training : bool, default=False
        If ``False``, fit a new classifier at each iteration. If ``True``, the
        classifier fitted in the previous iteration is used for further training in
        subsequent iterations.

    random_state : int, RandomState instance, default=None
        Control the randomization of the algorithm.

        - If int, ``random_state`` is the seed used by the random number
          generator;
        - If ``RandomState`` instance, random_state is the random number
          generator;
        - If ``None``, the random number generator is the ``RandomState``
          instance used by ``np.random``.

    Attributes
    ----------
    acquisition_func_ : function
        Method used to calculate the classification uncertainty at each iteration.
    evaluation_metric_ : scorer
        Metric used to estimate the performance of the AL classifier at each iteration.
    classifier_ : estimator object
        The classifier used in the iterative process. It is the classifier fitted in the
        last iteration.
    metadata_ : dict
        Contains the performance estimations, classifiers, labeled pool mask and original
        dataset. Note: the labeled pool stored at iteration ``i`` is used as the training
        set at iteration ``i+1``.
    n_init_ : int
        Number of observations included in the initial training dataset.
    budget_ : int
        Number of observations to be added to the training set per iteration. Also known
        as budget.
    max_iter_ : int
        Maximum number of iterations allowed.
    labeled_pool_ : array-like of shape (n_samples,)
        Mask that filters the labeled observations from the original dataset.
    """

    def __init__(
        self,
        classifier: Union[BaseEstimator, ClassifierMixin] = None,
        acquisition_func=None,
        n_init: Union[int, float] = None,
        budget: Union[int, float] = None,
        max_iter: int = None,
        evaluation_metric=None,
        continue_training: bool = False,
        random_state: int = None,
    ):
        self.classifier = classifier
        self.acquisition_func = acquisition_func
        self.n_init = n_init
        self.budget = budget
        self.max_iter = max_iter
        self.evaluation_metric = evaluation_metric
        self.continue_training = continue_training
        self.random_state = random_state

    @abstractmethod
    def _initialization(self, X, y=None, initial_selection=None):
        pass

    @abstractmethod
    def _iteration(self, X_lab, y_lab, X_pool=None, y_pool=None):
        pass

    @abstractmethod
    def _oracle(self, probabilities):
        pass

    def _early_stop(self):
        # more stopping conditions may be added here in the future
        return self.labeled_pool_.all()

    def _check(self, X, y):
        """Check initialization parameters to run an AL model."""

        # Classifier
        if self.classifier is None:
            self._classifier = RandomForestClassifier(random_state=self.random_state)
        else:
            self._classifier = clone(self.classifier)

        # Random state
        if (
            hasattr(self._classifier, "random_state")
            and self._classifier.random_state is None
            and self.random_state is not None
        ):
            self._classifier.set_params(random_state=self.random_state)

        # Acquisition function
        if self.acquisition_func is None:
            self.acquisition_func_ = ACQUISITION_FUNCTIONS["random"]
        elif type(self.acquisition_func) == str:
            self.acquisition_func_ = ACQUISITION_FUNCTIONS[self.acquisition_func]
        else:
            self.acquisition_func_ = self.acquisition_func

        # Number of initial observations
        if self.n_init is None or self.n_init < 1:
            perc = 0.02 if self.n_init is None else self.n_init
            n_init_ = int(np.round(perc * X.shape[0]))
        else:
            n_init_ = self.n_init
        self.n_init_ = n_init_ if n_init_ >= 2 else 2

        # Budget
        if self.budget is None or self.budget < 1:
            perc = 0.02 if self.budget is None else self.budget
            budget_ = int(np.round(perc * X.shape[0]))
        else:
            budget_ = self.budget
        self.budget_ = budget_ if budget_ >= 1 else 1

        # Maximum iterations
        self.max_iter_ = (
            self.max_iter
            if self.max_iter is not None
            else int(np.round((X.shape[0] - self.n_init_) / self.budget_))
        )

        # Evaluation metric
        if self.evaluation_metric is None:
            self.evaluation_metric_ = get_scorer("accuracy")
        elif type(self.evaluation_metric) == str:
            self.evaluation_metric_ = get_scorer(self.evaluation_metric)
        else:
            self.evaluation_metric_ = self.evaluation_metric

        # Train a different classifier per iteration (as an alternative continue training
        # the classifier from previous iterations)
        if type(self.continue_training) == bool:
            self._continue_training = self.continue_training
        else:
            raise TypeError(
                "``continue_training`` must be of type ``bool``. Got"
                f" {self.continue_training} instead."
            )

        # Set up basic attributes
        if not hasattr(self, "_current_iter"):
            self._current_iter = 0
            self.metadata_ = {
                "data": (X, y),
                "performance_metric": self.evaluation_metric_._score_func.__name__,
            }
        else:
            # Raise error, model is already fitted
            raise StopIteration(
                f"Active Learning model {type(self).__name__} is already initialized."
            )

    def _save_metadata(self, X, y, X_test=None, y_test=None):
        self.metadata_[self._current_iter] = {"labeled_pool": self.labeled_pool_.copy()}

        # Save performance in the training set
        if hasattr(self, "classifier_"):
            self.metadata_[self._current_iter]["train_score"] = self.evaluation_metric_(
                self.classifier_, X[self.labeled_pool_], y[self.labeled_pool_]
            )

        # Save performance in the test set
        if hasattr(self, "classifier_") and self._has_test:
            self.metadata_[self._current_iter]["test_score"] = self.evaluation_metric_(
                self.classifier_, X_test, y_test
            )

        # Save classifier
        if hasattr(self, "classifier_") and not self._continue_training:
            self.metadata_[self._current_iter]["classifier"] = self.classifier_

    def initialization(self, X, y, initial_selection=None, **kwargs):
        self._check(X, y)
        self.labeled_pool_ = self._initialization(
            X=X, y=y, initial_selection=initial_selection
        )
        self._save_metadata(X[self.labeled_pool_], y[self.labeled_pool_], **kwargs)
        return self

    def iteration(self, X, y, **kwargs):
        if not hasattr(self, "_current_iter"):
            # Raise error, model is not initialized
            raise StopIteration(
                f"Active Learning model {type(self).__name__} is not initialized yet."
            )

        # Create iteration's classifier if the previous one is not going to be
        # used
        if not self._continue_training:
            self.classifier_ = clone(self._classifier)

        # Run iteration
        probabilities = self._iteration(X, y, **kwargs)

        # Update labeled pool
        self._oracle(probabilities)

        # Save results from iteration
        self._current_iter += 1
        self._save_metadata(X, y, **kwargs)

        return self

    def fit(self, X, y, **kwargs):
        """
        Fit an Active Learning model from training set (X, y).

        Parameters
        ----------
        X : {array-like, sparse matrix} of shape (n_samples, n_features)
            The training input samples.

        y : array-like of shape (n_samples,) or (n_samples, n_outputs)
            The target values (class labels) as integers or strings.

        Returns
        -------
        self : Active Learning Classifier
            Fitted Active Learning model.
        """

        # Check if parameters are properly set up
        X, y = check_X_y(X, y)

        # Check if there is a Test set
        self._has_test = True if "X_test" in kwargs else False

        # Initialize the AL model
        self.initialization(X, y, **kwargs)

        # Create base classifier if it is going to be trained across iterations
        if self._continue_training and not hasattr(self, "classifier_"):
            self.classifier_ = clone(self._classifier)

        # Iterate
        for iter_ in range(1, self.max_iter_ + 1):

            self.iteration(X, y, **kwargs)

            if self._early_stop():
                break

        return self

    def predict(self, X):
        """
        Predict class or regression value for X.

        For a classification model, the predicted class for each sample in X is
        returned. For a regression model, the predicted value based on X is
        returned.

        Parameters
        ----------
        X : {array-like, sparse matrix} of shape (n_samples, n_features)
            The test input samples.

        Returns
        -------
        y : array-like of shape (n_samples,) or (n_samples, n_outputs)
            The predicted classes, or the predict values.
        """
        if not self._continue_training and self._has_test:
            iter_perf = np.array(
                [
                    [i, self.metadata_[i]["test_score"]]
                    for i in range(1, self._current_iter + 1)
                ]
            )
            index = iter_perf[np.argmax(iter_perf, axis=0)[1]][0]
            return self.metadata_[index]["classifier"].predict(X)
        else:
            return self.classifier_.predict(X)



from typing import Union
from copy import deepcopy
import numpy as np
from sklearn.base import ClassifierMixin, BaseEstimator, clone
from sklearn.model_selection import GridSearchCV
from imblearn.pipeline import Pipeline
from imblearn.over_sampling.base import BaseOverSampler

from .base import BaseActiveLearner


def _random_initialization(self, X=None, y=None, initial_selection=None):
    """Randomly select an initial training dataset."""
    if initial_selection is not None:
        labeled_pool = initial_selection
    else:
        rng = np.random.RandomState(self.random_state)
        labeled_pool = np.zeros(shape=(X.shape[0])).astype(bool)
        ids = rng.choice(np.arange(X.shape[0]), self.n_init_, replace=False)
        if np.unique(y[ids]).shape[0] == 1:
            ids[-1] = rng.choice(
                np.arange(X.shape[0])[y != y[ids][0]], 1, replace=False
            )
        labeled_pool[ids] = True
    return labeled_pool


class StandardAL(BaseActiveLearner, ClassifierMixin):
    """
    Standard Active Learning model with a random initial data selection

    Parameters
    ----------
    classifier : classifier object, default=None
        Classifier or pipeline to be trained in the iterative process. If None, defaults
        to sklearn's RandomForestClassifier with default parameters and uses the
        ``random_state`` passed in the Active Learning model.

    acquisition_func : function or {'entropy', 'breaking_ties',\
        'random'}, default=None
        Method used to quantify the prediction's uncertainty level. All predefined
        functions are set up so that a higher value means higher uncertainty (higher
        likelihood of selection) and vice-versa. The uncertainty estimate is used to
        select the instances to be added to the labeled/training dataset. Acquisition
        functions may be added or changed in the ``UNCERTAINTY_FUNCTIONS`` dictionary.
        If None, defaults to "random".

    n_init : int or float, default=None
        Number of observations to include in the initial training dataset. If
        ``n_init`` < 1, then the corresponding percentage of the original dataset
        will be used as the initial training set. If None, defaults to 2% of the size of
        the original dataset.

    budget : int or float, default=None
        Number of observations to be added to the training dataset at each iteration. If
        ``budget`` < 1, then the corresponding percentage of the original dataset will be
        used as the initial training set. If None, defaults to 2% of the size of the
        original dataset.

    max_iter : int, default=None
        Maximum number of iterations allowed. If None, the experiment will run until 100%
        of the dataset is added to the training set.

    evaluation_metric : string, default='accuracy'
        Metric used to calculate the test scores. See
        ``research.metrics`` for info on available
        performance metrics.

    continue_training : bool, default=False
        If ``False``, fit a new classifier at each iteration. If ``True``, the
        classifier fitted in the previous iteration is used for further training in
        subsequent iterations.

    random_state : int, RandomState instance, default=None
        Control the randomization of the algorithm.

        - If int, ``random_state`` is the seed used by the random number
          generator;
        - If ``RandomState`` instance, random_state is the random number
          generator;
        - If ``None``, the random number generator is the ``RandomState``
          instance used by ``np.random``.

    Attributes
    ----------
    acquisition_func_ : function
        Method used to calculate the classification uncertainty at each iteration.
    evaluation_metric_ : scorer
        Metric used to estimate the performance of the AL classifier at each iteration.
    classifier_ : estimator object
        The classifier used in the iterative process. It is the classifier fitted in the
        last iteration.
    metadata_ : dict
        Contains the performance estimations, classifiers, labeled pool mask and original
        dataset.
    n_init_ : int
        Number of observations included in the initial training dataset.
    budget_ : int
        Number of observations to be added to the training set per iteration. Also known
        as budget.
    max_iter_ : int
        Maximum number of iterations allowed.
    labeled_pool_ : array-like of shape (n_samples,)
        Mask that filters the labeled observations from the original dataset.
    """

    def _initialization(self, X=None, y=None, initial_selection=None):
        labeled_pool = _random_initialization(self, X, y, initial_selection)
        return labeled_pool

    def _iteration(self, X, y, **kwargs):
        self.classifier_.fit(X[self.labeled_pool_], y[self.labeled_pool_])
        return self.classifier_.predict_proba(X[~self.labeled_pool_])

    def _oracle(self, probabilities):
        uncertainties = self.acquisition_func_(probabilities)
        unlabeled_ids = np.argwhere(~self.labeled_pool_).squeeze()

        ids = (
            unlabeled_ids[np.argsort(uncertainties)[::-1][: self.budget_]]
            if unlabeled_ids.ndim >= 1
            else unlabeled_ids.flatten()[0]
        )

        self.labeled_pool_[ids] = True
        return self


class AugmentationAL(BaseActiveLearner, ClassifierMixin):
    """
    Active Learning with pipelined Data Augmentation. This method is implemented and
    analysed in a working paper.

    Parameters
    ----------
    classifier : classifier object, default=None
        Classifier or pipeline to be trained in the iterative process. If None, defaults
        to sklearn's RandomForestClassifier with default parameters and uses the
        ``random_state`` passed in the Active Learning model.

    generator : generator estimator, default=None
        Generator to be used for artificial data generation within Active
        Learning iterations.

    param_grid : dict or list of dictionaries
        Used to optimize the classifier and generator hyperparameters at each iteration
        via cross-validated grid-search. If None, parameter tuning is skipped.
        Dictionary with parameters names (``str``) as keys and lists of parameter
        settings to try as values, or a list of such dictionaries, in which case the
        grids spanned by each dictionary in the list are explored. This enables searching
        over any sequence of parameter settings.

    cv : int, cross-validation generator or an iterable, default=None
        Determines the cross-validation splitting strategy. Used to optimize the
        classifier and generator hyperparameters at each iteration.
        Possible inputs for cv are:

        - None, to use the default 5-fold cross validation,
        - integer, to specify the number of folds in a `(Stratified)KFold`,
        - `CV splitter`.

        For integer/None inputs, if the estimator is a classifier and ``y`` is
        either binary or multiclass, :class:`StratifiedKFold` is used. In all
        other cases, :class:`KFold` is used. These splitters are instantiated
        with `shuffle=False` so the splits will be the same across calls.

    acquisition_func : function or {'entropy', 'breaking_ties', 'random'}, default=None
        Method used to quantify the prediction's uncertainty level. All predefined
        functions are set up so that a higher value means higher uncertainty (higher
        likelihood of selection) and vice-versa. The uncertainty estimate is used to
        select the instances to be added to the labeled/training dataset. Acquisition
        functions may be added or changed in the ``UNCERTAINTY_FUNCTIONS`` dictionary.
        If None, defaults to "random".

    n_init : int or float, default=None
        Number of observations to include in the initial training dataset. If
        ``n_init`` < 1, then the corresponding percentage of the original dataset
        will be used as the initial training set. If None, defaults to 2% of the size of
        the original dataset.

    budget : int or float, default=None
        Number of observations to be added to the training dataset at each iteration. If
        ``budget`` < 1, then the corresponding percentage of the original dataset will be
        used as the initial training set. If None, defaults to 2% of the size of the
        original dataset.

    max_iter : int, default=None
        Maximum number of iterations allowed. If None, the experiment will run until 100%
        of the dataset is added to the training set.

    evaluation_metric : string, default='accuracy'
        Metric used to calculate the test scores. See
        ``mlresearch.metrics`` for info on available
        performance metrics.

    continue_training : bool, default=False
        If ``False``, fit a new classifier at each iteration. If ``True``, the
        classifier fitted in the previous iteration is used for further training in
        subsequent iterations.

    random_state : int, RandomState instance, default=None
        Control the randomization of the algorithm.

        - If int, ``random_state`` is the seed used by the random number generator;
        - If ``RandomState`` instance, random_state is the random number generator;
        - If ``None``, the random number generator is the ``RandomState`` instance used
          by ``np.random``.

    Attributes
    ----------
    acquisition_func_ : function
        Method used to calculate the classification uncertainty at each iteration.
    evaluation_metric_ : scorer
        Metric used to estimate the performance of the AL classifier at each iteration.
    classifier_ : estimator object
        The classifier used in the iterative process. It is the classifier fitted in the
        last iteration.
    metadata_ : dict
        Contains the performance estimations, classifiers, labeled pool mask and original
        dataset.
    n_init_ : int
        Number of observations included in the initial training dataset.
    budget_ : int
        Number of observations to be added to the training set per iteration.
    max_iter_ : int
        Maximum number of iterations allowed.
    labeled_pool_ : array-like of shape (n_samples,)
        Mask that filters the labeled observations from the original dataset.
    """

    def __init__(
        self,
        classifier: Union[BaseEstimator, ClassifierMixin] = None,
        generator: BaseOverSampler = None,
        param_grid: dict = None,
        cv=None,
        acquisition_func=None,
        n_init: Union[int, float] = None,
        budget: Union[int, float] = None,
        max_iter: int = None,
        evaluation_metric=None,
        continue_training: bool = False,
        random_state: int = None,
    ):
        super().__init__(
            classifier=classifier,
            acquisition_func=acquisition_func,
            n_init=n_init,
            budget=budget,
            max_iter=max_iter,
            evaluation_metric=evaluation_metric,
            continue_training=continue_training,
            random_state=random_state,
        )
        self.generator = generator
        self.param_grid = param_grid
        self.cv = cv

    def _check(self, X, y):
        super()._check(X, y)

        # Generator
        if (
            self.generator is not None
            and hasattr(self.generator, "random_state")
            and self.generator.random_state is None
            and self.random_state is not None
        ):
            # Check random state
            self._generator = clone(self.generator)
            self._generator.set_params(random_state=self.random_state)

            # Add generator to classifier as a pipeline
            generator = clone(self._generator)
            classifier = clone(self._classifier)
            self._classifier = Pipeline(
                [("generator", generator), ("classifier", classifier)]
            )

        # Check if parameters in param_grid are valid
        if type(self.param_grid) == dict:
            for key in self.param_grid.keys():
                if key not in self._classifier.get_params():
                    raise ValueError(
                        f"Invalid parameter {key} for generator or classifier in {self} "
                        "check the list of available parameters with "
                        "`almodel._classifier.get_params().keys()`."
                    )
        elif self.param_grid is not None:
            raise TypeError(
                f"``param_grid`` must be a dict or None. Got {self.param_grid} instead."
            )

    def _save_metadata(self, X, y, **kwargs):
        super()._save_metadata(X, y, **kwargs)
        if hasattr(self, "classifier_") and type(self.classifier_) == GridSearchCV:
            self.metadata_[self._current_iter]["parameters"] = {
                k: v
                for k, v in self.classifier_.best_estimator_.get_params().items()
                if k in self.param_grid.keys()
            }

    def _check_cross_validation(self, y):
        """Define cross-validation object"""

        min_frequency = np.unique(y, return_counts=True)[-1].min()
        cv = deepcopy(self.cv)

        if hasattr(self.cv, "n_splits"):
            cv.n_splits = min(min_frequency, cv.n_splits)
        elif type(self.cv) == int:
            cv = min(min_frequency, cv)
        elif cv is None:
            cv = min(min_frequency, 5)
        else:
            raise TypeError(
                "``cv`` object must be of type int or cross-validation generator. Got "
                f"{self.cv} instead"
            )
        return cv

    def _initialization(self, X=None, y=None, initial_selection=None):
        labeled_pool = _random_initialization(self, X, y, initial_selection)
        return labeled_pool

    def _iteration(self, X, y, **kwargs):

        # Set up parameter tuning within iterations
        cv = self._check_cross_validation(y[self.labeled_pool_])
        if self.param_grid is not None and cv != 1:
            self.classifier_ = GridSearchCV(
                estimator=self.classifier_,
                param_grid=self.param_grid,
                scoring=self.evaluation_metric_,
                cv=cv,
                refit=True,
            )
        self.classifier_.fit(X[self.labeled_pool_], y[self.labeled_pool_])
        return self.classifier_.predict_proba(X[~self.labeled_pool_])

    def _oracle(self, probabilities):
        uncertainties = self.acquisition_func_(probabilities)
        unlabeled_ids = np.argwhere(~self.labeled_pool_).squeeze()

        ids = (
            unlabeled_ids[np.argsort(uncertainties)[::-1][: self.budget_]]
            if unlabeled_ids.ndim >= 1
            else unlabeled_ids.flatten()[0]
        )

        self.labeled_pool_[ids] = True
        return self



"""
Download, transform and simulate various datasets.
"""

# Author: Joao Fonseca <jpfonseca@novaims.unl.pt>
# License: MIT

import os
from os.path import join
from urllib.parse import urljoin
from string import ascii_lowercase

from rich.progress import track
import numpy as np
import pandas as pd

from .base import Datasets, get_data_home, FETCH_URLS


class ContinuousCategoricalDatasets(Datasets):
    """Class to download, transform and save datasets with both continuous
    and categorical features."""

    @staticmethod
    def _modify_columns(data, categorical_features):
        """Rename and reorder columns of dataframe."""
        X_metric, X_cat, y = (
            data.drop(columns=categorical_features + ["target"]),
            data[categorical_features],
            data.target,
        )
        X_metric.columns = range(len(X_metric.columns))
        X_cat.columns = [f"cat_{i}" for i in range(len(X_cat.columns))]
        return pd.concat([X_metric, X_cat, y], axis=1)

    def download(self):
        """Download the datasets."""
        self.data_home_ = get_data_home(data_home=self.data_home)
        dataset_prefix = self.__class__.__name__.lower().replace("datasets", "")

        if self.names == "all":
            func_names = [func_name for func_name in dir(self) if "fetch_" in func_name]
        else:
            func_names = [
                f"fetch_{name}".lower().replace(" ", "_") for name in self.names
            ]
        self.content_ = []
        for func_name in track(func_names, description="Datasets"):
            dat_name = func_name.replace("fetch_", "")
            name = dat_name.upper().replace("_", " ")
            file_name = f"{dataset_prefix}_{dat_name}.csv"

            if (
                file_name not in os.listdir(self.data_home_)
                and self.download_if_missing
            ):
                df, categorical_features = getattr(self, func_name)()
                df = self._modify_columns(df, list(categorical_features))
                df.to_csv(join(self.data_home_, file_name), index=False)

            data = pd.read_csv(join(self.data_home_, file_name))

            self.content_.append((name, data))
        return self

    def summarize_datasets(self):
        """
        Create a summary of the downloaded datasets.

        Returns
        -------
        datasets_summary : pd.DataFrame
            Dataframe with summary statistics of all datasets.
        """

        datasets_summary = super(
            ContinuousCategoricalDatasets, self
        ).summarize_datasets()
        columns = datasets_summary.columns.tolist()

        # Define summary table columns and empty list
        summary_columns = ["Metric", "Non-Metric"]
        extended_summary = []

        # Populate empty list
        for name, dataset in self.content_:
            dataset = dataset.drop(columns="target")
            values = [
                sum(~dataset.columns.str.startswith("cat_")),
                sum(dataset.columns.str.startswith("cat_")),
            ]
            extended_summary.append(values)
        extended_summary = pd.DataFrame(extended_summary, columns=summary_columns)
        datasets_summary = pd.concat([datasets_summary, extended_summary], axis=1)

        # Reorder columns
        index = columns.index("Features") + 1
        columns = [*columns[:index], *summary_columns, *columns[index:]]
        return datasets_summary[columns]

    def fetch_adult(self):
        """Download and transform the Adult Data Set.

        https://archive.ics.uci.edu/ml/datasets/Adult
        """
        data = pd.read_csv(FETCH_URLS["adult"], header=None, na_values=" ?").dropna()
        data.rename(columns={data.columns[-1]: "target"}, inplace=True)

        mapper = {v: k for k, v in enumerate(data.target.unique())}
        data.target = data.target.map(mapper)

        categorical_features = [1, 3, 5, 6, 7, 8, 9, 13]
        return data, categorical_features

    def fetch_abalone(self):
        """Download and transform the Abalone Data Set.

        https://archive.ics.uci.edu/ml/datasets/Abalone
        """
        data = pd.read_csv(FETCH_URLS["abalone"], header=None)
        data.rename(columns={data.columns[-1]: "target"}, inplace=True)
        categorical_features = [0]
        return data, categorical_features

    def fetch_acute(self):
        """Download and transform the Acute Inflammations Data Set.

        https://archive.ics.uci.edu/ml/datasets/Acute+Inflammations
        """
        data = pd.read_csv(
            FETCH_URLS["acute"], header=None, sep="\t", decimal=",", encoding="UTF-16"
        )
        data["target"] = data[6].str[0] + data[7].str[0]
        data.drop(columns=[6, 7], inplace=True)

        mapper = {v: k for k, v in enumerate(data.target.unique())}
        data.target = data.target.map(mapper)

        categorical_features = list(range(1, 6))
        return data, categorical_features

    def fetch_annealing(self):
        """Download and transform the Annealing Data Set.

        https://archive.ics.uci.edu/ml/datasets/Annealing
        """
        data = pd.read_csv(FETCH_URLS["annealing"], header=None, na_values="?")

        # some features are dropped; they have too many missing values
        missing_feats = (data.isnull().sum(0) / data.shape[0]) < 0.1
        data = data.iloc[:, missing_feats.values]
        data[2].fillna(data[2].mode().squeeze(), inplace=True)

        data = data.T.reset_index(drop=True).T
        data.rename(columns={data.columns[-1]: "target"}, inplace=True)

        mapper = {v: k for k, v in enumerate(data.target.unique())}
        data.target = data.target.map(mapper)

        categorical_features = [0, 1, 5, 9]
        return data, categorical_features

    def fetch_census(self):
        """Download and transform the Census-Income (KDD) Data Set.

        https://archive.ics.uci.edu/ml/datasets/Census-Income+%28KDD%29
        """
        data = pd.read_csv(FETCH_URLS["census"], header=None)

        categorical_features = (
            list(range(1, 5))
            + list(range(6, 16))
            + list(range(19, 29))
            + list(range(30, 38))
            + [39]
        )

        # some features are dropped; they have too many missing values
        cols_ids = [1, 6, 9, 13, 14, 20, 21, 29, 31, 37]
        categorical_features = np.argwhere(
            np.delete(
                data.rename(columns={k: f"nom_{k}" for k in categorical_features})
                .columns.astype("str")
                .str.startswith("nom_"),
                cols_ids,
            )
        ).squeeze()
        data = data.drop(columns=cols_ids).T.reset_index(drop=True).T
        # some rows are dropped; they have rare missing values
        data = data.iloc[
            data.applymap(lambda x: x != " Not in universe").all(1).values, :
        ]

        data.rename(columns={data.columns[-1]: "target"}, inplace=True)

        mapper = {v: k for k, v in enumerate(data.target.unique())}
        data.target = data.target.map(mapper)

        return data, categorical_features

    def fetch_contraceptive(self):
        """Download and transform the Contraceptive Method Choice Data Set.

        https://archive.ics.uci.edu/ml/datasets/Contraceptive+Method+Choice
        """
        data = pd.read_csv(FETCH_URLS["contraceptive"], header=None)
        data.rename(columns={data.columns[-1]: "target"}, inplace=True)
        categorical_features = [4, 5, 6, 8]
        return data, categorical_features

    def fetch_covertype(self):
        """Download and transform the Covertype Data Set.

        https://archive.ics.uci.edu/ml/datasets/Covertype
        """
        data = pd.read_csv(FETCH_URLS["covertype"], header=None)
        data.rename(columns={data.columns[-1]: "target"}, inplace=True)
        wilderness_area = pd.Series(
            np.argmax(data.iloc[:, 10:14].values, axis=1), name=10
        )
        soil_type = pd.Series(np.argmax(data.iloc[:, 14:54].values, axis=1), name=11)
        data = (
            data.drop(columns=list(range(10, 54)))
            .join(wilderness_area)
            .join(soil_type)[list(range(0, 12)) + ["target"]]
        )
        categorical_features = [10, 11]
        return data, categorical_features

    def fetch_credit_approval(self):
        """Download and transform the Credit Approval Data Set.

        https://archive.ics.uci.edu/ml/datasets/Credit+Approval
        """
        data = pd.read_csv(
            FETCH_URLS["credit_approval"], header=None, na_values="?"
        ).dropna()
        data.rename(columns={data.columns[-1]: "target"}, inplace=True)

        mapper = {v: k for k, v in enumerate(data.target.unique())}
        data.target = data.target.map(mapper)

        categorical_features = [0, 3, 4, 5, 6, 8, 9, 11, 12]
        return data, categorical_features

    def fetch_dermatology(self):
        """Download and transform the Dermatology Data Set.

        https://archive.ics.uci.edu/ml/datasets/Dermatology
        """
        data = pd.read_csv(
            FETCH_URLS["dermatology"], header=None, na_values="?"
        ).dropna()
        data.rename(columns={data.columns[-1]: "target"}, inplace=True)
        categorical_features = list(range(data.shape[1] - 1))
        categorical_features.remove(33)
        return data, categorical_features

    def fetch_echocardiogram(self):
        """Download and transform the Echocardiogram Data Set.

        https://archive.ics.uci.edu/ml/datasets/Echocardiogram
        """
        data = pd.read_csv(
            FETCH_URLS["echocardiogram"],
            header=None,
            on_bad_lines="skip",
            na_values="?",
        )
        data.drop(columns=[10, 11], inplace=True)
        data.dropna(inplace=True)
        data.rename(columns={data.columns[-1]: "target"}, inplace=True)
        categorical_features = [1, 3]
        return data, categorical_features

    def fetch_flags(self):
        """Download and transform the Flags Data Set.

        https://archive.ics.uci.edu/ml/datasets/Flags
        """
        data = pd.read_csv(FETCH_URLS["flags"], header=None)
        target = data[6].rename("target")
        data = data.drop(columns=[0, 6]).T.reset_index(drop=True).T.join(target)
        categorical_features = [
            0,
            1,
            4,
            8,
            9,
            10,
            11,
            12,
            13,
            14,
            15,
            21,
            22,
            23,
            24,
            25,
            26,
            27,
        ]
        return data, categorical_features

    def fetch_heart_disease(self):
        """Download and transform the Heart Disease Data Set.

        https://archive.ics.uci.edu/ml/datasets/Heart+Disease
        """
        data = (
            pd.concat(
                [
                    pd.read_csv(url, header=None, na_values="?")
                    for url in FETCH_URLS["heart_disease"]
                ],
                ignore_index=True,
            )
            .drop(columns=[10, 11, 12])
            .dropna()
        )
        data.rename(columns={data.columns[-1]: "target"}, inplace=True)
        categorical_features = [1, 2, 5, 6, 8]
        return data, categorical_features

    def fetch_hepatitis(self):
        """Download and transform the Hepatitis Data Set.

        https://archive.ics.uci.edu/ml/datasets/Hepatitis
        """
        data = (
            pd.read_csv(FETCH_URLS["hepatitis"], header=None, na_values="?")
            .drop(columns=[15, 18])
            .dropna()
        )
        target = data[0].rename("target")
        data = data.drop(columns=[0]).T.reset_index(drop=True).T.join(target)
        categorical_features = list(range(1, 13)) + [16]
        return data, categorical_features

    def fetch_german_credit(self):
        """Download and transform the German Credit Data Set.

        https://archive.ics.uci.edu/ml/datasets/Statlog+%28German+Credit+Data%29
        """
        data = pd.read_csv(FETCH_URLS["german_credit"], header=None, sep=" ")
        data.rename(columns={data.columns[-1]: "target"}, inplace=True)
        categorical_features = (
            np.argwhere(data.iloc[0, :-1].apply(lambda x: str(x)[0] == "A").values)
            .squeeze()
            .tolist()
        )
        return data, categorical_features

    def fetch_heart(self):
        """Download and transform the Heart Data Set.

        http://archive.ics.uci.edu/ml/datasets/statlog+(heart)
        """
        data = pd.read_csv(FETCH_URLS["heart"], header=None, delim_whitespace=True)
        data.rename(columns={data.columns[-1]: "target"}, inplace=True)
        categorical_features = [1, 2, 5, 6, 8, 10, 12]
        return data, categorical_features

    def fetch_thyroid(self):
        """Download and transform the Thyroid Disease Data Set.
        Label 0 corresponds to no disease found.
        Label 1 corresponds to one or multiple diseases found.

        https://archive.ics.uci.edu/ml/datasets/Thyroid+Disease
        """
        data = (
            pd.read_csv(FETCH_URLS["thyroid"], header=None, na_values="?")
            .drop(columns=27)
            .dropna()
            .T.reset_index(drop=True)
            .T
        )
        data.rename(columns={data.columns[-1]: "target"}, inplace=True)
        data["target"] = (
            data["target"].apply(lambda x: x.split("[")[0]) != "-"
        ).astype(int)
        categorical_features = [
            1,
            2,
            3,
            4,
            5,
            6,
            7,
            8,
            9,
            10,
            11,
            12,
            13,
            14,
            15,
            16,
            18,
            20,
            22,
            24,
            26,
            27,
        ]
        return data, categorical_features


class MultiClassDatasets(Datasets):
    """Class to download, transform and save multi-class datasets."""

    def fetch_first_order_theorem(self):
        """Download and transform the First Order Theorem Data Set.

        https://www.openml.org/d/1475
        """
        data = pd.read_csv(FETCH_URLS["first_order_theorem"])
        data.rename(columns={"Class": "target"}, inplace=True)
        return data

    def fetch_gas_drift(self):
        """Download and transform the Gas Drift Data Set.

        https://www.openml.org/d/1476
        """
        data = pd.read_csv(FETCH_URLS["gas_drift"])
        data.rename(columns={"Class": "target"}, inplace=True)
        return data

    def fetch_autouniv_au7(self):
        """Download and transform the AutoUniv au7 Data Set

        https://www.openml.org/d/1552
        """
        data = pd.read_csv(FETCH_URLS["autouniv_au7"])
        data.rename(columns={"Class": "target"}, inplace=True)
        data.target = data.target.apply(lambda x: x.replace("class", "")).astype(int)

        mask = (data.iloc[:, :-1].nunique() > 10).tolist()
        mask.append(True)
        data = data.loc[:, mask].copy()
        return data

    def fetch_autouniv_au4(self):
        """Download and transform the AutoUniv au4 Data Set

        https://www.openml.org/d/1548
        """
        data = pd.read_csv(FETCH_URLS["autouniv_au4"])
        data.rename(columns={"Class": "target"}, inplace=True)
        data.target = data.target.apply(lambda x: x.replace("class", "")).astype(int)

        mask = (data.iloc[:, :-1].nunique() > 10).tolist()
        mask.append(True)
        data = data.loc[:, mask].copy()
        return data

    def fetch_mice_protein(self):
        """Download and transform the Mice Protein Data Set

        https://www.openml.org/d/40966
        """
        data = pd.read_csv(FETCH_URLS["mice_protein"])
        data.rename(columns={"class": "target"}, inplace=True)
        data.drop(columns=["MouseID"], inplace=True)
        data.replace("?", np.nan, inplace=True)

        mask = (data.iloc[:, :-1].nunique() > 10).tolist()
        mask.append(True)
        mask2 = data.isna().sum() < 10
        data = data.loc[:, mask & mask2].dropna().copy()

        data.iloc[:, :-1] = data.iloc[:, :-1].astype(float)

        mapper = {v: k for k, v in enumerate(data.target.unique())}
        data.target = data.target.map(mapper)
        return data

    def fetch_steel_plates(self):
        """Download and transform the Steel Plates Fault Data Set.

        https://www.openml.org/d/40982
        """
        data = pd.read_csv(FETCH_URLS["steel_plates"])

        mask = (data.iloc[:, :-1].nunique() > 10).tolist()
        mask.append(True)
        data = data.loc[:, mask].copy()

        mapper = {v: k for k, v in enumerate(data.target.unique())}
        data.target = data.target.map(mapper)
        return data

    def fetch_cardiotocography(self):
        """Download and transform the Cardiotocography Data Set.

        https://www.openml.org/d/1560
        """
        data = pd.read_csv(FETCH_URLS["cardiotocography"])
        data.rename(columns={"Class": "target"}, inplace=True)

        mask = (data.iloc[:, :-1].nunique() > 10).tolist()
        mask.append(True)
        data = data.loc[:, mask].copy()

        return data

    def fetch_waveform(self):
        """Download and transform the Waveform Database Generator (version 2) Data Set.

        https://www.openml.org/d/60
        """
        data = pd.read_csv(FETCH_URLS["waveform"])
        data.rename(columns={"class": "target"}, inplace=True)
        return data

    def fetch_volkert(self):
        """Download and transform the Volkert Data Set.

        https://www.openml.org/d/41166
        """
        data = pd.read_csv(FETCH_URLS["volkert"])
        data.rename(columns={"class": "target"}, inplace=True)

        mask = (data.iloc[:, 1:].nunique() > 100).tolist()
        mask.insert(0, True)
        data = data.loc[:, mask].copy()
        return data

    def fetch_vehicle(self):
        """Download and transform the Vehicle Silhouettes Data Set.

        https://archive.ics.uci.edu/ml/datasets/Statlog+(Vehicle+Silhouettes)
        """
        data = pd.DataFrame()
        for letter in ascii_lowercase[0:9]:
            partial_data = pd.read_csv(
                urljoin(FETCH_URLS["vehicle"], "xa%s.dat" % letter),
                header=None,
                delim_whitespace=True,
            )
            partial_data = partial_data.rename(columns={18: "target"})
            data = data.append(partial_data)

        mapper = {v: k for k, v in enumerate(data.target.unique())}
        data.target = data.target.map(mapper)

        return data

    def fetch_asp_potassco(self):
        """Download and transform the ASP-POTASSCO Data Set.

        https://www.openml.org/d/41705
        """
        data = pd.read_csv(FETCH_URLS["asp_potassco"], na_values="?")
        data.dropna(inplace=True)
        data["target"] = data["algorithm"]
        data.drop(columns=["instance_id", "algorithm"], inplace=True)

        mask = (data.iloc[:, :-1].nunique() > 100).tolist()
        mask.append(True)
        data = data.loc[:, mask].copy()

        mapper = {v: k for k, v in enumerate(data.target.unique())}
        data.target = data.target.map(mapper)

        return data

    def fetch_wine_quality(self):
        """Download and transform the Wine Quality Data Set.

        https://www.openml.org/d/40691
        """
        data = pd.read_csv(FETCH_URLS["wine_quality"])
        data.rename(columns={"class": "target"}, inplace=True)
        return data

    def fetch_mfeat_zernike(self):
        """Download and transform the Multiple Features Dataset: Zernike Data Set.

        https://www.openml.org/d/22
        """
        data = pd.read_csv(FETCH_URLS["mfeat_zernike"])
        data.drop_duplicates(inplace=True)
        data.rename(columns={"class": "target"}, inplace=True)
        return data

    def fetch_gesture_segmentation(self):
        """Download and transform the Gesture Phase Segmentation Data Set.

        https://www.openml.org/d/4538
        """
        data = pd.read_csv(FETCH_URLS["gesture_segmentation"])
        data.rename(columns={"Phase": "target"}, inplace=True)

        mapper = {v: k for k, v in enumerate(data.target.unique())}
        data.target = data.target.map(mapper)
        return data

    def fetch_texture(self):
        """Download and transform the Texture Data Set.

        https://www.openml.org/d/40499
        """
        data = pd.read_csv(FETCH_URLS["texture"])
        data.drop_duplicates(inplace=True)
        data.rename(columns={"Class": "target"}, inplace=True)
        return data

    def fetch_usps(self):
        """Download and transform the USPS Data Set.

        https://www.openml.org/data/get_csv/19329737/usps.arff
        """
        data = pd.read_csv(FETCH_URLS["usps"])
        data.rename(columns={"int0": "target"}, inplace=True)
        return data

    def fetch_vowels(self):
        """Download and transform the Vowels Data Set.

        https://www.openml.org/d/375
        """
        data = pd.read_csv(FETCH_URLS["vowels"])
        data.rename(columns={"speaker": "target"}, inplace=True)
        data.drop(columns=["utterance", "frame"], inplace=True)
        return data

    def fetch_pendigits(self):
        """Download and transform the Pen-Based Recognition of Handwritten
        Digits Data Set.

        https://www.openml.org/d/32
        """
        data = pd.read_csv(FETCH_URLS["pendigits"])
        data.rename(columns={"class": "target"}, inplace=True)
        return data

    def fetch_image_segmentation(self):
        """Download and transform the Image Segmentation Data Set.

        https://www.openml.org/d/40984
        """
        data = pd.read_csv(FETCH_URLS["image_segmentation"])
        data.drop(columns=data.columns[:5], inplace=True)
        data.rename(columns={"class": "target"}, inplace=True)

        mapper = {v: k for k, v in enumerate(data.target.unique())}
        data.target = data.target.map(mapper)
        return data

    def fetch_baseball(self):
        """Download and transform the Baseball Hall of Fame Data Set.

        https://www.openml.org/d/185
        """
        data = pd.read_csv(FETCH_URLS["baseball"], na_values="?")
        data.drop(columns=["Player", "Position"], inplace=True)
        data.rename(columns={"Hall_of_Fame": "target"}, inplace=True)
        data.dropna(inplace=True)
        return data



from os.path import join, dirname, abspath


def generate_paths(filepath):
    """
    Generate data, results and analysis paths.
    """
    prefix_path = join(dirname(abspath(filepath)), "..")
    paths = [join(prefix_path, name) for name in ("data", "results", "analysis")]
    return paths



"""
Acquisition functions that may be used along with Active Learning objects. All functions
must be set up so that a higher value means higher uncertainty (higher likelihood of
selection) and vice-versa.
"""
import numpy as np


def breaking_ties(probabilities):
    """Breaking Ties uncertainty measurement. The output is scaled and reversed."""
    probs_sorted = np.sort(probabilities, axis=1)[:, ::-1]
    # The extra minus is redundant but I kept as a highlight of the change in the
    # original formula.
    bt = -(probs_sorted[:, 0] - probs_sorted[:, 1])
    return bt


def entropy(probabilities):
    """Shannon's Entropy-based uncertainty measurement."""
    return (-probabilities * np.log2(probabilities)).sum(axis=1)


def random(probabilities):
    """Random data selection. The uncertainty is the same for all observations."""
    return np.ones(probabilities.shape[0]) / probabilities.shape[0]


ACQUISITION_FUNCTIONS = {
    "entropy": entropy,
    "breaking_ties": breaking_ties,
    "random": random,
}



"""
Contains several functions to prepare and format tables for LaTeX documents.
"""

# Author: Joao Fonseca <jpmrfonseca@gmail.com>
# License: MIT

from itertools import product
import numpy as np
import pandas as pd


def _check_indices(table_index, indices):
    """Formats the ``indices`` parameter passed."""

    if indices is None:
        indices = table_index.to_frame().to_dict("list")

    # Check indices - list
    if type(indices) == list:
        if all([type(i) in [str, int] for i in indices]):
            indices_ = {table_index.names[0]: {i: i for i in indices}}
        elif all([type(i) == list for i in indices]):
            indices_ = {
                name: {i: i for i in values}
                for name, values in zip(table_index.names, indices)
            }
        elif all([type(i) == dict for i in indices]):
            indices_ = {
                name: values for name, values in zip(table_index.names, indices)
            }

    # Check indices - dict
    if type(indices) == dict:
        if all([type(i) == str for i in list(indices.values())]):
            index_name = table_index.names[0] if table_index.names[0] is not None else 0
            indices_ = {index_name: indices}
        elif all([type(i) == list for i in list(indices.values())]):
            indices_ = {
                name: {i: i for i in values} for name, values in indices.items()
            }
        elif all([type(i) == dict for i in list(indices.values())]):
            indices_ = indices

    # Replace index name if there's only a single index with no name
    if len(indices_) == 1 and None in indices_.keys():
        indices_[0] = indices_[None]
        indices_.pop(None)

    return indices_


def format_table(table, indices=None, columns=None, drop_missing=True):
    """
    Sort and rename rows and columns. Mostly used to set results from
    experiments in the intended order.

    Arguments
    ---------
    table : pd.DataFrame
        Dataframe with results statistics. Optionally, it may be indexed with related
        metadata. Supports both single and multi index.

    indices : list, dict
        Iterable with index names and order of index. Can be used to order index columns,
        rows and filter unwanted rows or columns:

        - If list of strings, orders index according to the passed values.
        - If list of lists, orders multi indices according to the passed values.
        - If list of dicts, orders and renames indices according to the passed values.
        - If dict, orders and renames index values.
        - If dict of lists orders index columns and values.
        - If dict of dicts orders both index columns and values and renames values.

    columns : list, dict
        Iterable with column names and order of columns. Can be used to order columns,
        filter unwanted columns:

        - If list, orders and selects columns
        - If dict, orders, selects and renames columns

    drop_missing : bool, default=True
        If True, removes any values, columns or indices not included in ``indices`` and
        ``columns``. If False, the columns and indices not included will be moved to the
        end of the dataframe.

    Returns
    -------
    formatted_table : pd.DataFrame
        Formatted dataframe.

    """

    table = table.copy()
    index_dtype = table.index.dtype
    index_name = table.index.names

    # Check indices
    indices_ = _check_indices(table.index, indices)

    # Get index data
    id_data = table.index.to_frame().reset_index(drop=True)
    id_order = list(indices_.keys())

    # Update mappers with missing values if necessary
    if not drop_missing:
        id_order = id_order + id_data.columns[~id_data.columns.isin(id_order)].tolist()
        indices_ = {
            key: {
                **mapper,
                **{i: i for i in id_data[key].unique() if i not in mapper.keys()},
            }
            for key, mapper in indices_.items()
        }

    # Map index values
    for key, cat in indices_.items():
        id_data[key] = pd.Categorical(id_data[key].map(cat), categories=cat.values())

    # Sort index cols and replace index values
    id_data = id_data[id_order]
    id_ = pd.MultiIndex.from_arrays(id_data.values.T, names=id_data.columns)
    table.set_index(id_, inplace=True)

    # Sort index vals
    index_vals_unique = [list(mapper.values()) for mapper in indices_.values()]
    index_values_sorted = (
        list(product(*index_vals_unique))
        if len(index_vals_unique) > 1
        else index_vals_unique[0]
    )
    table = table.loc[index_values_sorted]
    table.index = table.index.astype(index_dtype)
    if len(index_name) == 1 and index_name[0] is None:
        table.index.name = None

    # Update column list/dictionary if necessary
    if type(columns) == list:
        columns_ = {i: i for i in columns}
    elif type(columns) == dict:
        columns_ = columns

    if not drop_missing:
        missing = {i: i for i in table.columns if i not in columns_.keys()}
        columns_ = {**columns_, **missing}

    # Sort columns
    formatted_table = table[list(columns_.keys())].rename(columns=columns_).copy()

    return formatted_table


def _make_bold(row, maximum=True, decimals=2, threshold=None, with_sem=False):
    """
    Make bold the lowest or highest value(s).
    with_sem returns an incomplete textbf latex function and a mask array.

    This function should be used in a pandas dataframe using the .apply method.
    """
    row = round(row, decimals)
    if threshold is None:
        val = row.max() if maximum else row.min()
        mask = row == val
    else:
        mask = (row > threshold) if maximum else (row < threshold)

    formatter = "{0:.%sf}" % decimals
    row = row.apply(lambda el: formatter.format(el))

    row[mask] = [
        "\\textbf{%s" % v if with_sem else "\\textbf{%s}" % v for v in row[mask]
    ]

    # Return mask only if function is being used to generate
    # a table with sem values
    if with_sem:
        return row, mask
    else:
        return row


def make_bold(table, maximum=True, threshold=None, decimals=2, axis=1):
    """
    Make bold the lowest or highest values, or values lower than, or higher than the
    passed value in ``threshold`` per row or column.

    Arguments
    ---------
    table : pd.DataFrame
        Dataframe with results statistics. Must not contain non-indexed metadata.
        Supports both single and multi index.

    maximum : bool, default=True
        Whether to look for the highest or lowest values:

        - If True and ``threshold`` is None, boldfaces the highest value in each
          row/column.
        - If False and ``threshold`` is None, boldfaces the lowest value in each
          row/column.
        - If True and ``threshold`` is not None, boldfaces all values above the given
          threshold.
        - If False and ``threshold`` is not None, boldfaces all values below the given
          threshold.

    threshold : int or float, default=None
        Threshold to boldface values. If None, one value will be boldfaced per row or
        column. If not None, boldfaces all values above or below ``threshold``.

    decimals : int, default=2
        Number of decimal places to round each value to.

    axis : {0 or 'index', 1 or 'columns'}, default=1
        Axis along which the function is applied:

        - 0 or 'index': apply function to column.
        - 1 or 'columns': apply function to each row.

    """
    return table.apply(
        lambda row: _make_bold(
            row, maximum=maximum, decimals=decimals, threshold=threshold
        ),
        axis=axis,
    )


def make_mean_sem_table(
    mean_vals,
    sem_vals=None,
    make_bold=False,
    maximum=True,
    threshold=None,
    decimals=2,
    axis=1,
):
    """
    Generate table with rounded decimals, bold maximum/minimum values or values
    above/below a given threshold, and combine mean and sem values.

    Arguments
    ---------
    mean_vals : pd.DataFrame
        Dataframe with results statistics. Must not
        contain non-indexed metadata. Supports both single and multi index.

    sem_vals : {pd.DataFrame or np.ndarray}, default=None
        Dataframe with standard errors of the means. If it is a DataFrame, must not
        contain non-indexed metadata. Supports both single and multi index.

    make_bold : bool, default=False
        If True, make bold the lowest or highest values, or values lower than, or higher
        than the passed value in ``threshold`` per row or column. If False, the
        parameters ``maximum``, ``threshold`` and ``axis`` are ignored.

    maximum : bool, default=True
        Whether to look for the highest or lowest values:

        - If True and ``threshold`` is None, boldfaces the highest value in each
          row/column.
        - If False and ``threshold`` is None, boldfaces the lowest value in each
          row/column.
        - If True and ``threshold`` is not None, boldfaces all values above the given
          threshold.
        - If False and ``threshold`` is not None, boldfaces all values below the given
          threshold.

    threshold : int or float, default=None
        Threshold to boldface values. If None, one value will be boldfaced per row or
        column. If not None, boldfaces all values above or below ``threshold``.

    decimals : int, default=2
        Number of decimal places to round each value to.

    axis : {0 or 'index', 1 or 'columns'}, default=1
        Axis along which the function is applied:

        - 0 or 'index': apply function to column.
        - 1 or 'columns': apply function to each row.

    Returns
    -------
    scores : pd.DataFrame
        Dataframe with the specified formatting.
    """

    if sem_vals is not None:

        if type(sem_vals) == np.ndarray:
            sem_vals = pd.DataFrame(
                sem_vals, index=mean_vals.index, columns=mean_vals.columns
            )

        scores = (
            mean_vals.applymap(("{:,.%sf}" % decimals).format)
            + r" $\pm$ "
            + sem_vals.applymap(("{:,.%sf}" % decimals).format)
        )
    else:
        scores = mean_vals.applymap(("{:,.%sf}" % decimals).format)

    if make_bold:
        mask = mean_vals.apply(
            lambda row: _make_bold(row, maximum, decimals, threshold, with_sem=True)[1],
            axis=axis,
        ).values

        scores.iloc[:, :] = np.where(mask, "\\textbf{" + scores + "}", scores)

    return scores


def export_longtable(table, path=None, caption=None, label=None, index=True):
    """
    Exports a pandas dataframe to LaTeX (longtable) format.

    This function replaces ``df.to_latex`` when there are latex commands in
    the table. Requires ``\\usepackage{booktabs}`` and ``\\usepackage{longtable}`` in the
    LaTeX preamble.

    .. warning::
        Do not pass a table with column names, index names or values containing an
        underscore ("_"), as it will result in an error in LaTeX. If you wish to have
        underscores in the LaTeX table, add a backslash before the underscore.

    Arguments
    ---------
    table : pd.DataFrame
        Dataframe with results statistics. Must not contain non-indexed metadata.
        Supports both single and multi index.

    path : str, default=None
        File path to write to. If None, the output is returned as a string.

    caption : str or tuple, default=None
        Tuple (full_caption, short_caption), which results in
        ``\\caption[short_caption]{full_caption}``; if a single string is passed, no
        short caption will be set.

    label : str, default=None
        The LaTeX label to be placed inside ``\\label{}`` in the output. This is used
        with ``\\ref{}`` in the main ``.tex`` file.

    index : bool, default=True
        Write row names (index).

    Returns
    -------
    tex_table : str or None
        If ``path`` is None, returns the result as a string. Otherwise returns None.
    """

    tex_table = (
        table.to_latex(
            longtable=True,
            caption=caption,
            label=label,
            index=index,
            column_format="c" * table.shape[1],
        )
        .replace(r"\textbackslash ", "\\")
        .replace(r"\{", "{")
        .replace(r"\}", "}")
        .replace(r"\$", "$")
    )

    if path is not None:
        open(path, "w").write(tex_table)
    else:
        return tex_table



"""
This module contains several functions to prepare and format tables for LaTeX
documents.
"""

from ._utils import (
    format_table,
    make_bold,
    make_mean_sem_table,
    export_longtable,
)


__all__ = [
    "format_table",
    "make_bold",
    "make_mean_sem_table",
    "export_longtable",
]



from urllib.request import urlopen
import multiprocessing.dummy as mp
from multiprocessing import cpu_count
import ssl
import numpy as np
import pandas as pd

from ..base import Datasets, FETCH_URLS
from .._multiclass import ContinuousCategoricalDatasets

ssl._create_default_https_context = ssl._create_unverified_context

MIN_CLASS = 20
MAJ_CLASS = 100
N_FEATURES = 4
X = pd.DataFrame(
    np.random.random((30 + MIN_CLASS + MAJ_CLASS, N_FEATURES)),
    columns=[str(i) for i in range(N_FEATURES)],
)
y = pd.Series(
    [
        *[0 for i in range(MIN_CLASS)],
        *[1 for i in range(30)],
        *[2 for i in range(MAJ_CLASS)],
    ],
    name="target",
)


def test_urls():
    """Test whether URLS are working."""
    urls = [
        url
        for sublist in [[url] for url in list(FETCH_URLS.values()) if type(url) == str]
        for url in sublist
    ]

    p = mp.Pool(cpu_count())
    url_status = p.map(lambda url: (urlopen(url).status == 200), urls)

    assert all(url_status)


def test_imbalance_datasets():
    """Test if the imbalance_datasets and summarize_datasets functions are working."""
    base_ir = MAJ_CLASS / MIN_CLASS
    irs = [1, 5, 10, 100]
    exp_irs = [i for i in irs if i > base_ir]

    name = "test"
    data = pd.concat([X, y], axis=1)

    datasets = Datasets()
    datasets.content_ = [(name, data)]
    for ir in irs:
        datasets.imbalance_datasets(ir)

    descr = datasets.summarize_datasets()
    content = datasets.content_

    # imbalance_datasets
    assert len(content) == len(exp_irs) + 1
    assert list(dict(content).keys()) == [name] + [f"{name} ({ir})" for ir in exp_irs]

    # summarize_datasets
    assert descr.shape == (len(exp_irs) + 1, 7)
    assert (descr["Features"] == N_FEATURES).all()
    assert descr["Dataset name"].tolist() == list(dict(content).keys())
    assert descr["Imbalance Ratio"].astype(int).tolist() == [
        i for i in irs if i >= base_ir
    ]


def test_categorical_summary():
    """Test if modification in summarize_datasets for ContinuousCategoricalDatasets
    is working."""
    base_ir = MAJ_CLASS / MIN_CLASS
    irs = [1, 5, 10, 100]
    exp_irs = [i for i in irs if i > base_ir]

    datasets = ContinuousCategoricalDatasets()
    datasets.content_ = [("test", pd.concat([X, y], axis=1))]

    # Set up mixed type datasets
    for n_cat in range(1, N_FEATURES + 1):
        cat_columns = {
            col: f"cat_{i}"
            for i, col in enumerate(np.random.choice(X.columns, n_cat, replace=False))
        }
        name = f"test_{n_cat}"
        data = pd.concat([X.rename(columns=cat_columns), y], axis=1)
        data.loc[:, cat_columns.values()] = (
            data[cat_columns.values()].round().astype(int)
        )
        datasets.content_.append((name, data))

    # Imbalance datasets
    for ir in irs:
        datasets.imbalance_datasets(ir)

    descr = datasets.summarize_datasets()

    # summarize_datasets
    assert descr.shape == ((len(exp_irs) + 1) * (N_FEATURES + 1), 9)
    assert (descr["Features"] == N_FEATURES).all()
    assert descr["Dataset name"].tolist() == list(dict(datasets.content_).keys())
    assert (descr["Metric"] + descr["Non-Metric"] == descr["Features"]).all()

    exp_cat = (
        descr["Dataset name"]
        .apply(lambda x: x.split("_")[-1].split(" ")[0])
        .replace({"test": 0})
        .astype(int)
    )
    true_cat = descr["Non-Metric"]
    assert (exp_cat == true_cat).all()
